@@ -96,12 +96,14 @@ RSpec.describe "bundler/inline#gemfile" do
   it "lets me use my own ui object" do
     script <<-RUBY, :artifice => "endpoint"
       require '#{entrypoint}'
-      class MyBundlerUI < Bundler::UI::Silent
+      class MyBundlerUI < Bundler::UI::Shell
         def confirm(msg, newline = nil)
           puts "CONFIRMED!"
         end
       end
-      gemfile(true, :ui => MyBundlerUI.new) do
+      my_ui = MyBundlerUI.new
+      my_ui.level = "confirm"
+      gemfile(true, :ui => my_ui) do
         source "https://notaserver.com"
         gem "activesupport", :require => true
       end
@@ -451,6 +453,9 @@ RSpec.describe "bundler/inline#gemfile" do
 
     realworld_system_gems "timeout uri" # this spec uses net/http which requires these default gems
 
+    # on prerelease rubies, a required_rubygems_version constraint is added by RubyGems to the resolution, causing Molinillo to load the `set` gem
+    realworld_system_gems "set --version 1.0.3" if Gem.ruby_version.prerelease?
+
     script <<-RUBY, :dir => tmp("path_without_gemfile"), :env => { "BUNDLER_GEM_DEFAULT_DIR" => system_gem_path.to_s }
       require "bundler/inline"
 
