@@ -1,43 +1,28 @@
 # frozen_string_literal: true
 
-require 'rubygems'
-
-# If bundler gemspec exists, add to stubs
-bundler_gemspec = File.expand_path("../../../bundler/bundler.gemspec", __FILE__)
-if File.exist?(bundler_gemspec)
-  Gem::Specification.dirs.unshift File.dirname(bundler_gemspec)
-  Gem::Specification.class_variable_set :@@stubs, nil
-  Gem::Specification.stubs
-  Gem::Specification.dirs.shift
-end
+require "rubygems"
 
 begin
-  gem 'test-unit', '~> 3.0'
+  gem "test-unit", "~> 3.0"
 rescue Gem::LoadError
 end
 
-if File.exist?(bundler_gemspec)
-  require_relative '../../bundler/lib/bundler'
-else
-  require 'bundler'
-end
-
-require 'test/unit'
+require "test/unit"
 
 ENV["JARS_SKIP"] = "true" if Gem.java_platform? # avoid unnecessary and noisy `jar-dependencies` post install hook
 
-require 'rubygems/deprecate'
+require "rubygems/deprecate"
 
-require 'fileutils'
-require 'pathname'
-require 'pp'
-require 'rubygems/package'
-require 'shellwords'
-require 'tmpdir'
-require 'uri'
-require 'zlib'
-require 'benchmark' # stdlib
-require 'rubygems/mock_gem_ui'
+require "fileutils"
+require "pathname"
+require "pp"
+require "rubygems/package"
+require "shellwords"
+require "tmpdir"
+require "uri"
+require "zlib"
+require "benchmark" # stdlib
+require "rubygems/mock_gem_ui"
 
 module Gem
 
@@ -119,6 +104,12 @@ class Gem::TestCase < Test::Unit::TestCase
     assert File.directory?(path), msg
   end
 
+  def refute_directory_exists(path, msg = nil)
+    msg = build_message(msg, "Expected path '#{path}' not to be a directory")
+    assert_path_not_exist path
+    refute File.directory?(path), msg
+  end
+
   # https://github.com/seattlerb/minitest/blob/21d9e804b63c619f602f3f4ece6c71b48974707a/lib/minitest/assertions.rb#L188
   def _synchronize
     yield
@@ -159,15 +150,15 @@ class Gem::TestCase < Test::Unit::TestCase
   # the original value when the block ends
 
   def enable_shared(value)
-    enable_shared = RbConfig::CONFIG['ENABLE_SHARED']
-    RbConfig::CONFIG['ENABLE_SHARED'] = value
+    enable_shared = RbConfig::CONFIG["ENABLE_SHARED"]
+    RbConfig::CONFIG["ENABLE_SHARED"] = value
 
     yield
   ensure
     if enable_shared
-      RbConfig::CONFIG['enable_shared'] = enable_shared
+      RbConfig::CONFIG["ENABLE_SHARED"] = enable_shared
     else
-      RbConfig::CONFIG.delete 'enable_shared'
+      RbConfig::CONFIG.delete "ENABLE_SHARED"
     end
   end
 
@@ -176,20 +167,20 @@ class Gem::TestCase < Test::Unit::TestCase
   # original value when the block ends
   #
   def vendordir(value)
-    vendordir = RbConfig::CONFIG['vendordir']
+    vendordir = RbConfig::CONFIG["vendordir"]
 
     if value
-      RbConfig::CONFIG['vendordir'] = value
+      RbConfig::CONFIG["vendordir"] = value
     else
-      RbConfig::CONFIG.delete 'vendordir'
+      RbConfig::CONFIG.delete "vendordir"
     end
 
     yield
   ensure
     if vendordir
-      RbConfig::CONFIG['vendordir'] = vendordir
+      RbConfig::CONFIG["vendordir"] = vendordir
     else
-      RbConfig::CONFIG.delete 'vendordir'
+      RbConfig::CONFIG.delete "vendordir"
     end
   end
 
@@ -199,21 +190,21 @@ class Gem::TestCase < Test::Unit::TestCase
   #
   def bindir(value)
     with_clean_path_to_ruby do
-      bindir = RbConfig::CONFIG['bindir']
+      bindir = RbConfig::CONFIG["bindir"]
 
       if value
-        RbConfig::CONFIG['bindir'] = value
+        RbConfig::CONFIG["bindir"] = value
       else
-        RbConfig::CONFIG.delete 'bindir'
+        RbConfig::CONFIG.delete "bindir"
       end
 
       begin
         yield
       ensure
         if bindir
-          RbConfig::CONFIG['bindir'] = bindir
+          RbConfig::CONFIG["bindir"] = bindir
         else
-          RbConfig::CONFIG.delete 'bindir'
+          RbConfig::CONFIG.delete "bindir"
         end
       end
     end
@@ -224,20 +215,20 @@ class Gem::TestCase < Test::Unit::TestCase
   # original value when the block ends
   #
   def exeext(value)
-    exeext = RbConfig::CONFIG['EXEEXT']
+    exeext = RbConfig::CONFIG["EXEEXT"]
 
     if value
-      RbConfig::CONFIG['EXEEXT'] = value
+      RbConfig::CONFIG["EXEEXT"] = value
     else
-      RbConfig::CONFIG.delete 'EXEEXT'
+      RbConfig::CONFIG.delete "EXEEXT"
     end
 
     yield
   ensure
     if exeext
-      RbConfig::CONFIG['EXEEXT'] = exeext
+      RbConfig::CONFIG["EXEEXT"] = exeext
     else
-      RbConfig::CONFIG.delete 'EXEEXT'
+      RbConfig::CONFIG.delete "EXEEXT"
     end
   end
 
@@ -258,23 +249,23 @@ class Gem::TestCase < Test::Unit::TestCase
       end
     end
 
-    targets << '' if targets.empty?
+    targets << "" if targets.empty?
 
     targets
   end
 
   def assert_contains_make_command(target, output, msg = nil)
-    if output.match(/\n/)
+    if output.include?("\n")
       msg = build_message(msg,
         "Expected output containing make command \"%s\", but was \n\nBEGIN_OF_OUTPUT\n%sEND_OF_OUTPUT" % [
-          ('%s %s' % [make_command, target]).rstrip,
+          ("%s %s" % [make_command, target]).rstrip,
           output,
         ]
       )
     else
       msg = build_message(msg,
         'Expected make command "%s", but was "%s"' % [
-          ('%s %s' % [make_command, target]).rstrip,
+          ("%s %s" % [make_command, target]).rstrip,
           output,
         ]
       )
@@ -302,18 +293,24 @@ class Gem::TestCase < Test::Unit::TestCase
   # or <tt>i686-darwin8.10.1</tt> otherwise.
 
   def setup
+    @orig_hooks = {}
     @orig_env = ENV.to_hash
     @tmp = File.expand_path("tmp")
 
     FileUtils.mkdir_p @tmp
 
-    ENV['GEM_VENDOR'] = nil
-    ENV['GEMRC'] = nil
-    ENV['XDG_CACHE_HOME'] = nil
-    ENV['XDG_CONFIG_HOME'] = nil
-    ENV['XDG_DATA_HOME'] = nil
-    ENV['SOURCE_DATE_EPOCH'] = nil
-    ENV['BUNDLER_VERSION'] = nil
+    @tempdir = Dir.mktmpdir("test_rubygems_", @tmp)
+    @tempdir.tap(&Gem::UNTAINT)
+
+    ENV["GEM_VENDOR"] = nil
+    ENV["GEMRC"] = nil
+    ENV["XDG_CACHE_HOME"] = nil
+    ENV["XDG_CONFIG_HOME"] = nil
+    ENV["XDG_DATA_HOME"] = nil
+    ENV["XDG_STATE_HOME"] = nil
+    ENV["SOURCE_DATE_EPOCH"] = nil
+    ENV["BUNDLER_VERSION"] = nil
+    ENV["RUBYGEMS_PREVENT_UPDATE_SUGGESTION"] = "true"
 
     @current_dir = Dir.pwd
     @fetcher     = nil
@@ -324,27 +321,23 @@ class Gem::TestCase < Test::Unit::TestCase
     # capture output
     Gem::DefaultUserInteraction.ui = Gem::MockGemUi.new
 
-    @tempdir = Dir.mktmpdir("test_rubygems_", @tmp)
-    @tempdir.tap(&Gem::UNTAINT)
-
-    ENV["TMPDIR"] = @tempdir
-
     @orig_SYSTEM_WIDE_CONFIG_FILE = Gem::ConfigFile::SYSTEM_WIDE_CONFIG_FILE
     Gem::ConfigFile.send :remove_const, :SYSTEM_WIDE_CONFIG_FILE
     Gem::ConfigFile.send :const_set, :SYSTEM_WIDE_CONFIG_FILE,
-                         File.join(@tempdir, 'system-gemrc')
+                         File.join(@tempdir, "system-gemrc")
 
-    @gemhome  = File.join @tempdir, 'gemhome'
-    @userhome = File.join @tempdir, 'userhome'
-    ENV["GEM_SPEC_CACHE"] = File.join @tempdir, 'spec_cache'
+    @gemhome  = File.join @tempdir, "gemhome"
+    @userhome = File.join @tempdir, "userhome"
+    @statehome = File.join @tempdir, "statehome"
+    ENV["GEM_SPEC_CACHE"] = File.join @tempdir, "spec_cache"
 
-    @orig_ruby = if ENV['RUBY']
+    @orig_ruby = if ENV["RUBY"]
       ruby = Gem.ruby
-      Gem.ruby = ENV['RUBY']
+      Gem.ruby = ENV["RUBY"]
       ruby
     end
 
-    @git = ENV['GIT'] || (win_platform? ? 'git.exe' : 'git')
+    @git = ENV["GIT"] || "git#{RbConfig::CONFIG['EXEEXT']}"
 
     Gem.ensure_gem_subdirectories @gemhome
     Gem.ensure_default_gem_subdirectories @gemhome
@@ -365,11 +358,12 @@ class Gem::TestCase < Test::Unit::TestCase
 
     Dir.chdir @tempdir
 
-    ENV['HOME'] = @userhome
+    ENV["HOME"] = @userhome
     Gem.instance_variable_set :@config_file, nil
     Gem.instance_variable_set :@user_home, nil
     Gem.instance_variable_set :@config_home, nil
     Gem.instance_variable_set :@data_home, nil
+    Gem.instance_variable_set :@state_home, @statehome
     Gem.instance_variable_set :@gemdeps, nil
     Gem.instance_variable_set :@env_requirements_by_name, nil
     Gem.send :remove_instance_variable, :@ruby_version if
@@ -377,12 +371,12 @@ class Gem::TestCase < Test::Unit::TestCase
 
     FileUtils.mkdir_p @userhome
 
-    ENV['GEM_PRIVATE_KEY_PASSPHRASE'] = PRIVATE_KEY_PASSPHRASE
+    ENV["GEM_PRIVATE_KEY_PASSPHRASE"] = PRIVATE_KEY_PASSPHRASE
 
     Gem.instance_variable_set(:@default_specifications_dir, nil)
     if Gem.java_platform?
-      @orig_default_gem_home = RbConfig::CONFIG['default_gem_home']
-      RbConfig::CONFIG['default_gem_home'] = @gemhome
+      @orig_default_gem_home = RbConfig::CONFIG["default_gem_home"]
+      RbConfig::CONFIG["default_gem_home"] = @gemhome
     else
       Gem.instance_variable_set(:@default_dir, @gemhome)
     end
@@ -404,7 +398,6 @@ class Gem::TestCase < Test::Unit::TestCase
     Gem.loaded_specs.clear
     Gem.instance_variable_set(:@activated_gem_paths, 0)
     Gem.clear_default_specs
-    Bundler.reset!
 
     Gem.configuration.verbose = true
     Gem.configuration.update_sources = true
@@ -418,15 +411,14 @@ class Gem::TestCase < Test::Unit::TestCase
     Gem.searcher = nil
     Gem::SpecFetcher.fetcher = nil
 
-    @orig_arch = RbConfig::CONFIG['arch']
+    @orig_arch = RbConfig::CONFIG["arch"]
 
     if win_platform?
-      util_set_arch 'i386-mswin32'
+      util_set_arch "i386-mswin32"
     else
-      util_set_arch 'i686-darwin8.10.1'
+      util_set_arch "i686-darwin8.10.1"
     end
 
-    @orig_hooks = {}
     %w[post_install_hooks done_installing_hooks post_uninstall_hooks pre_uninstall_hooks pre_install_hooks pre_reset_hooks post_reset_hooks post_build_hooks].each do |name|
       @orig_hooks[name] = Gem.send(name).dup
     end
@@ -451,7 +443,7 @@ class Gem::TestCase < Test::Unit::TestCase
       end
     end
 
-    RbConfig::CONFIG['arch'] = @orig_arch
+    RbConfig::CONFIG["arch"] = @orig_arch
 
     if defined? Gem::RemoteFetcher
       Gem::RemoteFetcher.fetcher = nil
@@ -461,7 +453,7 @@ class Gem::TestCase < Test::Unit::TestCase
 
     FileUtils.rm_rf @tempdir
 
-    ENV.replace(@orig_env)
+    restore_env
 
     Gem::ConfigFile.send :remove_const, :SYSTEM_WIDE_CONFIG_FILE
     Gem::ConfigFile.send :const_set, :SYSTEM_WIDE_CONFIG_FILE,
@@ -469,13 +461,13 @@ class Gem::TestCase < Test::Unit::TestCase
 
     Gem.ruby = @orig_ruby if @orig_ruby
 
-    RbConfig::CONFIG['mandir'] = @orig_mandir
-    RbConfig::CONFIG['sitelibdir'] = @orig_sitelibdir
-    RbConfig::CONFIG['bindir'] = @orig_bindir
+    RbConfig::CONFIG["mandir"] = @orig_mandir
+    RbConfig::CONFIG["sitelibdir"] = @orig_sitelibdir
+    RbConfig::CONFIG["bindir"] = @orig_bindir
 
     Gem.instance_variable_set :@default_specifications_dir, nil
     if Gem.java_platform?
-      RbConfig::CONFIG['default_gem_home'] = @orig_default_gem_home
+      RbConfig::CONFIG["default_gem_home"] = @orig_default_gem_home
     else
       Gem.instance_variable_set :@default_dir, nil
     end
@@ -491,9 +483,9 @@ class Gem::TestCase < Test::Unit::TestCase
   end
 
   def credential_setup
-    @temp_cred = File.join(@userhome, '.gem', 'credentials')
+    @temp_cred = File.join(@userhome, ".gem", "credentials")
     FileUtils.mkdir_p File.dirname(@temp_cred)
-    File.write @temp_cred, ':rubygems_api_key: 701229f217cdf23b1344c7b4b54ca97'
+    File.write @temp_cred, ":rubygems_api_key: 701229f217cdf23b1344c7b4b54ca97"
     File.chmod 0600, @temp_cred
   end
 
@@ -548,10 +540,10 @@ class Gem::TestCase < Test::Unit::TestCase
   #
   # Yields the +specification+ to the block, if given
 
-  def git_gem(name = 'a', version = 1)
+  def git_gem(name = "a", version = 1)
     have_git?
 
-    directory = File.join 'git', name
+    directory = File.join "git", name
     directory = File.expand_path directory
 
     git_spec = Gem::Specification.new name, version do |specification|
@@ -562,22 +554,23 @@ class Gem::TestCase < Test::Unit::TestCase
 
     gemspec = "#{name}.gemspec"
 
-    File.open File.join(directory, gemspec), 'w' do |io|
+    File.open File.join(directory, gemspec), "w" do |io|
       io.write git_spec.to_ruby
     end
 
     head = nil
 
     Dir.chdir directory do
-      unless File.exist? '.git'
-        system @git, 'init', '--quiet'
-        system @git, 'config', 'user.name',  'RubyGems Tests'
-        system @git, 'config', 'user.email', 'rubygems@example'
+      unless File.exist? ".git"
+        system @git, "init", "--quiet"
+        system @git, "checkout", "-b", "master", "--quiet"
+        system @git, "config", "user.name",  "RubyGems Tests"
+        system @git, "config", "user.email", "rubygems@example"
       end
 
-      system @git, 'add', gemspec
-      system @git, 'commit', '-a', '-m', 'a non-empty commit message', '--quiet'
-      head = Gem::Util.popen(@git, 'rev-parse', 'master').strip
+      system @git, "add", gemspec
+      system @git, "commit", "-a", "-m", "a non-empty commit message", "--quiet"
+      head = Gem::Util.popen(@git, "rev-parse", "master").strip
     end
 
     return name, git_spec.version, directory, head
@@ -589,13 +582,13 @@ class Gem::TestCase < Test::Unit::TestCase
   def have_git?
     return if in_path? @git
 
-    pend 'cannot find git executable, use GIT environment variable to set'
+    pend "cannot find git executable, use GIT environment variable to set"
   end
 
   def in_path?(executable) # :nodoc:
-    return true if %r{\A([A-Z]:|/)} =~ executable and File.exist? executable
+    return true if %r{\A([A-Z]:|/)} =~ executable && File.exist?(executable)
 
-    ENV['PATH'].split(File::PATH_SEPARATOR).any? do |directory|
+    ENV["PATH"].split(File::PATH_SEPARATOR).any? do |directory|
       File.exist? File.join directory, executable
     end
   end
@@ -604,7 +597,7 @@ class Gem::TestCase < Test::Unit::TestCase
   # Builds and installs the Gem::Specification +spec+
 
   def install_gem(spec, options = {})
-    require 'rubygems/installer'
+    require "rubygems/installer"
 
     gem = spec.cache_file
 
@@ -618,7 +611,7 @@ class Gem::TestCase < Test::Unit::TestCase
       gem = File.join(@tempdir, File.basename(gem)).tap(&Gem::UNTAINT)
     end
 
-    Gem::Installer.at(gem, options.merge({:wrappers => true})).install
+    Gem::Installer.at(gem, options.merge({ :wrappers => true })).install
   end
 
   ##
@@ -631,7 +624,7 @@ class Gem::TestCase < Test::Unit::TestCase
   ##
   # Uninstalls the Gem::Specification +spec+
   def uninstall_gem(spec)
-    require 'rubygems/uninstaller'
+    require "rubygems/uninstaller"
 
     Class.new(Gem::Uninstaller) do
       def ask_if_ok(spec)
@@ -654,7 +647,7 @@ class Gem::TestCase < Test::Unit::TestCase
   # Reads a Marshal file at +path+
 
   def read_cache(path)
-    File.open path.dup.tap(&Gem::UNTAINT), 'rb' do |io|
+    File.open path.dup.tap(&Gem::UNTAINT), "rb" do |io|
       Marshal.load io.read
     end
   end
@@ -674,7 +667,7 @@ class Gem::TestCase < Test::Unit::TestCase
     dir = File.dirname path
     FileUtils.mkdir_p dir unless File.directory? dir
 
-    File.open path, 'wb' do |io|
+    File.open path, "wb" do |io|
       yield io if block_given?
     end
 
@@ -717,16 +710,16 @@ class Gem::TestCase < Test::Unit::TestCase
   #
   # Use this with #write_file to build an installed gem.
 
-  def quick_gem(name, version='2')
-    require 'rubygems/specification'
+  def quick_gem(name, version="2")
+    require "rubygems/specification"
 
     spec = Gem::Specification.new do |s|
       s.platform    = Gem::Platform::RUBY
       s.name        = name
       s.version     = version
-      s.author      = 'A User'
-      s.email       = 'example@example.com'
-      s.homepage    = 'http://example.com'
+      s.author      = "A User"
+      s.email       = "example@example.com"
+      s.homepage    = "http://example.com"
       s.summary     = "this is a summary"
       s.description = "This is a test description"
 
@@ -757,7 +750,7 @@ class Gem::TestCase < Test::Unit::TestCase
         next if File.exist? file
         FileUtils.mkdir_p File.dirname(file)
 
-        File.open file, 'w' do |fp|
+        File.open file, "w" do |fp|
           fp.puts "# #{file}"
         end
       end
@@ -843,15 +836,15 @@ class Gem::TestCase < Test::Unit::TestCase
   # or a +block+ can be given for full customization of the specification.
 
   def util_spec(name, version = 2, deps = nil, *files) # :yields: specification
-    raise "deps or block, not both" if deps and block_given?
+    raise "deps or block, not both" if deps && block_given?
 
     spec = Gem::Specification.new do |s|
       s.platform    = Gem::Platform::RUBY
       s.name        = name
       s.version     = version
-      s.author      = 'A User'
-      s.email       = 'example@example.com'
-      s.homepage    = 'http://example.com'
+      s.author      = "A User"
+      s.email       = "example@example.com"
+      s.homepage    = "http://example.com"
       s.summary     = "this is a summary"
       s.description = "This is a test description"
 
@@ -862,7 +855,7 @@ class Gem::TestCase < Test::Unit::TestCase
 
     if deps
       deps.keys.each do |n|
-        spec.add_dependency n, (deps[n] || '>= 0')
+        spec.add_dependency n, (deps[n] || ">= 0")
       end
     end
 
@@ -889,7 +882,7 @@ class Gem::TestCase < Test::Unit::TestCase
     if deps
       block = proc do |s|
         deps.keys.each do |n|
-          s.add_dependency n, (deps[n] || '>= 0')
+          s.add_dependency n, (deps[n] || ">= 0")
         end
       end
     end
@@ -898,7 +891,7 @@ class Gem::TestCase < Test::Unit::TestCase
 
     util_build_gem spec
 
-    cache_file = File.join @tempdir, 'gems', "#{spec.original_name}.gem"
+    cache_file = File.join @tempdir, "gems", "#{spec.original_name}.gem"
     FileUtils.mkdir_p File.dirname cache_file
     FileUtils.mv spec.cache_file, cache_file
     FileUtils.rm spec.spec_file
@@ -940,11 +933,11 @@ class Gem::TestCase < Test::Unit::TestCase
   # TODO: nuke this and fix tests. this should speed up a lot
 
   def util_make_gems(prerelease = false)
-    @a1 = quick_gem 'a', '1' do |s|
+    @a1 = quick_gem "a", "1" do |s|
       s.files = %w[lib/code.rb]
       s.require_paths = %w[lib]
       s.date = Gem::Specification::TODAY - 86400
-      s.homepage = 'http://a.example.com'
+      s.homepage = "http://a.example.com"
       s.email = %w[example@example.com example2@example.com]
       s.authors = %w[Example Example2]
       s.description = <<-DESC
@@ -961,27 +954,27 @@ Also, a list:
       s.require_paths = %w[lib]
     end
 
-    @a2      = quick_gem('a', '2',      &init)
-    @a3a     = quick_gem('a', '3.a',    &init)
-    @a_evil9 = quick_gem('a_evil', '9', &init)
-    @b2      = quick_gem('b', '2',      &init)
-    @c1_2    = quick_gem('c', '1.2',    &init)
-    @x       = quick_gem('x', '1',      &init)
-    @dep_x   = quick_gem('dep_x', '1') do |s|
+    @a2      = quick_gem("a", "2",      &init)
+    @a3a     = quick_gem("a", "3.a",    &init)
+    @a_evil9 = quick_gem("a_evil", "9", &init)
+    @b2      = quick_gem("b", "2",      &init)
+    @c1_2    = quick_gem("c", "1.2",    &init)
+    @x       = quick_gem("x", "1",      &init)
+    @dep_x   = quick_gem("dep_x", "1") do |s|
       s.files = %w[lib/code.rb]
       s.require_paths = %w[lib]
-      s.add_dependency 'x', '>= 1'
+      s.add_dependency "x", ">= 1"
     end
 
-    @pl1 = quick_gem 'pl', '1' do |s| # l for legacy
+    @pl1 = quick_gem "pl", "1" do |s| # l for legacy
       s.files = %w[lib/code.rb]
       s.require_paths = %w[lib]
-      s.platform = Gem::Platform.new 'i386-linux'
-      s.instance_variable_set :@original_platform, 'i386-linux'
+      s.platform = Gem::Platform.new "i386-linux"
+      s.instance_variable_set :@original_platform, "i386-linux"
     end
 
     if prerelease
-      @a2_pre = quick_gem('a', '2.a', &init)
+      @a2_pre = quick_gem("a", "2.a", &init)
       write_file File.join(*%W[gems #{@a2_pre.original_name} lib code.rb])
       util_build_gem @a2_pre
     end
@@ -1007,7 +1000,7 @@ Also, a list:
   # Set the platform to +arch+
 
   def util_set_arch(arch)
-    RbConfig::CONFIG['arch'] = arch
+    RbConfig::CONFIG["arch"] = arch
     platform = Gem::Platform.new arch
 
     Gem.instance_variable_set :@platforms, nil
@@ -1098,7 +1091,7 @@ Also, a list:
     Zlib::Deflate.deflate data
   end
 
-  def util_set_RUBY_VERSION(version, patchlevel = nil, revision = nil, description = nil, engine = "ruby", engine_version = nil)
+  def util_set_RUBY_VERSION(version, patchlevel, revision, description, engine = "ruby", engine_version = nil)
     if Gem.instance_variables.include? :@ruby_version
       Gem.send :remove_instance_variable, :@ruby_version
     end
@@ -1106,16 +1099,16 @@ Also, a list:
     @RUBY_VERSION        = RUBY_VERSION
     @RUBY_PATCHLEVEL     = RUBY_PATCHLEVEL     if defined?(RUBY_PATCHLEVEL)
     @RUBY_REVISION       = RUBY_REVISION       if defined?(RUBY_REVISION)
-    @RUBY_DESCRIPTION    = RUBY_DESCRIPTION    if defined?(RUBY_DESCRIPTION)
+    @RUBY_DESCRIPTION    = RUBY_DESCRIPTION
     @RUBY_ENGINE         = RUBY_ENGINE
     @RUBY_ENGINE_VERSION = RUBY_ENGINE_VERSION if defined?(RUBY_ENGINE_VERSION)
 
     util_clear_RUBY_VERSION
 
     Object.const_set :RUBY_VERSION,        version
-    Object.const_set :RUBY_PATCHLEVEL,     patchlevel     if patchlevel
-    Object.const_set :RUBY_REVISION,       revision       if revision
-    Object.const_set :RUBY_DESCRIPTION,    description    if description
+    Object.const_set :RUBY_PATCHLEVEL,     patchlevel
+    Object.const_set :RUBY_REVISION,       revision
+    Object.const_set :RUBY_DESCRIPTION,    description
     Object.const_set :RUBY_ENGINE,         engine
     Object.const_set :RUBY_ENGINE_VERSION, engine_version if engine_version
   end
@@ -1128,8 +1121,7 @@ Also, a list:
       defined?(@RUBY_PATCHLEVEL)
     Object.const_set :RUBY_REVISION,       @RUBY_REVISION    if
       defined?(@RUBY_REVISION)
-    Object.const_set :RUBY_DESCRIPTION,    @RUBY_DESCRIPTION if
-      defined?(@RUBY_DESCRIPTION)
+    Object.const_set :RUBY_DESCRIPTION,    @RUBY_DESCRIPTION
     Object.const_set :RUBY_ENGINE,         @RUBY_ENGINE
     Object.const_set :RUBY_ENGINE_VERSION, @RUBY_ENGINE_VERSION if
       defined?(@RUBY_ENGINE_VERSION)
@@ -1177,7 +1169,7 @@ Also, a list:
   # Borland) versus Cygwin, Mingw, etc.
 
   def self.vc_windows?
-    RUBY_PLATFORM.match('mswin')
+    RUBY_PLATFORM.match("mswin")
   end
 
   ##
@@ -1185,7 +1177,15 @@ Also, a list:
   # Borland) versus Cygwin, Mingw, etc.
 
   def vc_windows?
-    RUBY_PLATFORM.match('mswin')
+    RUBY_PLATFORM.match("mswin")
+  end
+
+  ##
+  # Is this test being run on a ruby/ruby repository?
+  #
+
+  def ruby_repo?
+    !ENV["GEM_COMMAND"].nil?
   end
 
   ##
@@ -1194,7 +1194,7 @@ Also, a list:
   # other platforms, including Cygwin, it will return 'make'.
 
   def self.make_command
-    ENV["make"] || ENV["MAKE"] || (vc_windows? ? 'nmake' : 'make')
+    ENV["make"] || ENV["MAKE"] || (vc_windows? ? "nmake" : "make")
   end
 
   ##
@@ -1203,14 +1203,14 @@ Also, a list:
   # other platforms, including Cygwin, it will return 'make'.
 
   def make_command
-    ENV["make"] || ENV["MAKE"] || (vc_windows? ? 'nmake' : 'make')
+    ENV["make"] || ENV["MAKE"] || (vc_windows? ? "nmake" : "make")
   end
 
   ##
   # Returns whether or not the nmake command could be found.
 
   def nmake_found?
-    system('nmake /? 1>NUL 2>&1')
+    system("nmake /? 1>NUL 2>&1")
   end
 
   # In case we're building docs in a background process, this method waits for
@@ -1263,13 +1263,13 @@ Also, a list:
     ruby = ENV["RUBY"]
     return ruby if ruby
     ruby = "ruby"
-    rubyexe = "#{ruby}.exe"
+    rubyexe = "#{ruby}#{RbConfig::CONFIG['EXEEXT']}"
 
     3.times do
-      if File.exist? ruby and File.executable? ruby and !File.directory? ruby
+      if File.exist?(ruby) && File.executable?(ruby) && !File.directory?(ruby)
         return File.expand_path(ruby)
       end
-      if File.exist? rubyexe and File.executable? rubyexe
+      if File.exist?(rubyexe) && File.executable?(rubyexe)
         return File.expand_path(rubyexe)
       end
       ruby = File.join("..", ruby)
@@ -1287,7 +1287,11 @@ Also, a list:
   end
 
   def rubygems_path
-    $LOAD_PATH.find{|p| p == File.dirname($LOADED_FEATURES.find{|f| f.end_with?("/rubygems.rb") }) }
+    $LOAD_PATH.find {|p| p == File.dirname($LOADED_FEATURES.find {|f| f.end_with?("/rubygems.rb") }) }
+  end
+
+  def bundler_path
+    $LOAD_PATH.find {|p| p == File.dirname($LOADED_FEATURES.find {|f| f.end_with?("/bundler.rb") }) }
   end
 
   def with_clean_path_to_ruby
@@ -1418,8 +1422,8 @@ Also, a list:
   #
   # Yields the +specification+ to the block, if given
 
-  def vendor_gem(name = 'a', version = 1)
-    directory = File.join 'vendor', name
+  def vendor_gem(name = "a", version = 1)
+    directory = File.join "vendor", name
 
     FileUtils.mkdir_p directory
 
@@ -1432,12 +1436,12 @@ Also, a list:
   #
   # Yields the +specification+ to the block, if given
 
-  def save_gemspec(name = 'a', version = 1, directory = '.')
+  def save_gemspec(name = "a", version = 1, directory = ".")
     vendor_spec = Gem::Specification.new name, version do |specification|
       yield specification if block_given?
     end
 
-    File.open File.join(directory, "#{name}.gemspec"), 'w' do |io|
+    File.open File.join(directory, "#{name}.gemspec"), "w" do |io|
       io.write vendor_spec.to_ruby
     end
 
@@ -1551,25 +1555,42 @@ Also, a list:
   # :stopdoc:
   # only available in RubyGems tests
 
-  PRIVATE_KEY_PASSPHRASE = 'Foo bar'.freeze
+  PRIVATE_KEY_PASSPHRASE = "Foo bar".freeze
 
   begin
-    PRIVATE_KEY                 = load_key 'private'
-    PRIVATE_KEY_PATH            = key_path 'private'
+    PRIVATE_KEY                 = load_key "private"
+    PRIVATE_KEY_PATH            = key_path "private"
 
     # ENCRYPTED_PRIVATE_KEY is PRIVATE_KEY encrypted with PRIVATE_KEY_PASSPHRASE
-    ENCRYPTED_PRIVATE_KEY       = load_key 'encrypted_private', PRIVATE_KEY_PASSPHRASE
-    ENCRYPTED_PRIVATE_KEY_PATH  = key_path 'encrypted_private'
+    ENCRYPTED_PRIVATE_KEY       = load_key "encrypted_private", PRIVATE_KEY_PASSPHRASE
+    ENCRYPTED_PRIVATE_KEY_PATH  = key_path "encrypted_private"
 
     PUBLIC_KEY                  = PRIVATE_KEY.public_key
 
-    PUBLIC_CERT                 = load_cert 'public'
-    PUBLIC_CERT_PATH            = cert_path 'public'
+    PUBLIC_CERT                 = load_cert "public"
+    PUBLIC_CERT_PATH            = cert_path "public"
   rescue Errno::ENOENT
     PRIVATE_KEY = nil
     PUBLIC_KEY  = nil
     PUBLIC_CERT = nil
   end if Gem::HAVE_OPENSSL
+
+  private
+
+  def restore_env
+    unless Gem.win_platform?
+      ENV.replace(@orig_env)
+      return
+    end
+
+    # Fallback logic for Windows below to workaround
+    # https://bugs.ruby-lang.org/issues/16798. Can be dropped once all
+    # supported rubies include the fix for that.
+
+    ENV.clear
+
+    @orig_env.each {|k, v| ENV[k] = v }
+  end
 end
 
 # https://github.com/seattlerb/minitest/blob/13c48a03d84a2a87855a4de0c959f96800100357/lib/minitest/mock.rb#L192
@@ -1579,7 +1600,7 @@ class Object
 
     metaclass = class << self; self; end
 
-    if respond_to? name and not methods.map(&:to_s).include? name.to_s
+    if respond_to?(name) && !methods.map(&:to_s).include?(name.to_s)
       metaclass.send :define_method, name do |*args|
         super(*args)
       end
@@ -1606,4 +1627,4 @@ class Object
   end unless method_defined?(:stub) # lib/resolv/test_dns.rb also has the same method definition
 end
 
-require_relative 'utilities'
+require_relative "utilities"
